.section .text, "x" 
 
.global QualityMixer 
	 
	 
/******************************************************************************
 
#converts 4 8-bit signed samples in $reg1 into 4 16-bit signed samples across $reg1 and 
$reg2 
#i.e. before: $reg1 = [dcba] 
#after:  $reg1 = [b0a0], $reg2 = [d0c0] 
*/
 
.macro conv8bitToQual1 reg1, reg2, mask			@ $reg1 -> [dcba], $reg2 irrelevant 
		 
		mov		\reg2, \reg1, LSR #16			@ $reg1:[dcba] -> $reg2:[00dc] 
		mov		\reg2, \reg2, LSL #8			@ $reg2:[00dc] -> $reg2:[0dc0] 
		orr		\reg2, \reg2, \reg2, LSL #8		@ $reg2:[0dc0] -> $reg2:[dec0] (e = c|d) 
		and		\reg2, \reg2, \mask				@ $reg2:[dec0] -> $reg2:[d0c0] 
		 
		mov		\reg1, \reg1, LSL #16			@ $reg1:[dcba] -> $reg1:[ba00] 
		orr		\reg1, \reg1, \reg1, LSR #8		@ $reg1:[ba00] -> $reg1:[bea0] (e = a|b) 
		and		\reg1, \reg1, \mask				@ $reg1:[0ba0] -> $reg1:[b0a0] 
.endm 
    
 
/* *****************************************************************************
 
# converts 1 8-bit signed sample in $inp into 2 16-bit signed samples in $reg1 
# i.e. before: $inp = [dcba] 
# after:  $inp = [0dbc], $reg1 = [0a0a] 
*/
 
.macro conv8bitToQual2 inp, reg1					@ $reg1 -> [dcba], $reg2 irrelevant 
		 
		mov		\reg1, \inp, LSL #8					@ get first byte, convert to 16-bit 
		mov		\inp, \inp, LSR #8					@ prep for next sample by moving right 8 bits 
		and		\reg1, \reg1, #0xFF00				@ mask out lower 16-bits 
		orr		\reg1, \reg1, \reg1, LSL #16		@ make upper and lower word the same 
.endm 
 
 
/* *****************************************************************************
 
# converts 1 8-bit signed sample in $inp into 4 16-bit signed samples across $reg1, 
$reg2 
# i.e. before: $inp = [dcba] 
# after:  $inp = [0dbc], $reg1 = [0a0a], $reg2 = [0a0a] 
*/
 
.macro conv8bitToQual4 inp, reg1, reg2			@ $reg1 -> [dcba], $reg2 irrelevant 
		 
		conv8bitToQual2 \inp, \reg1 
		mov		\reg2, \reg1						@ now 2 32-bit i.e. 4 samples 
.endm 
 
 
/* *****************************************************************************
 
# QualityMixer(R0, R1, R2, R3) 
# 
# R0 = source buffer - 8bit signed 
# R1 = dest   buffer - 16bit signed 
# R2 = size (bytes source) 
# R3 = quality: 1, 2, 4 
# *****************************************************************************
*/
 
 
QualityMixer: 
	stmdb sp!, {r4-r12} 
 
	cmp		r3, #1 
	beq		QualityMixer1 
	cmp		r3, #2 
	beq		QualityMixer2 
	cmp		r3, #4 
	beq		QualityMixer4 
	b		QualityMixerExit 
	 
/* *****************************************************************************
*/
 
QualityMixer1: 
	mov		r3, r2					@ r3 = size 
	and		r3, r3, #15				@ r3 = size % 16 (effectively get remainder when divided by 16) 
	 
	subs	r2, r2, r3				@ r2 = r2 - (r2 % 16) i.e. number of 16's 
	beq		QualityMixer1Remainder 
 
	mov		r4, #0xFF00				@ r4 = mask 0x0000FF00 
	orr     r4, r4, r4, LSL #16		@ r4 = mask 0xFF00FF00 (to load 32-const) 
	 
QualityMixer1MainLoop: 
	/*pld		[r0]	 */
	ldmia	r0!, {r5, r7, r9, r11} 
	 
	conv8bitToQual1 r5, r6, r4 
	conv8bitToQual1 r7, r8, r4 
	conv8bitToQual1 r9, r10, r4 
	conv8bitToQual1 r11, r12, r4 
	 
	stmia	r1!, {r5-r12} 
 
	subs	r2, r2, #16 
	bne		QualityMixer1MainLoop 
	 
QualityMixer1Remainder: 
	teq		r3, #0 
	beq		QualityMixerExit 
 
QualityMixer1RemainderLoop: 
 
	ldrb	r4, [r0] 
	add		r0, r0, #1 
	mov		r4, r4, LSL #8 
	strh	r4, [r1] 
	add		r1, r1, #2 
	subs	r3, r3, #1 
	bne		QualityMixer1RemainderLoop 
	b		QualityMixerExit 
 
 
/* *****************************************************************************
*/
 
QualityMixer2: 
	cmp		r2, #8 
	blt		QualityMixer2Remainder	@ only remainder case 
 
QualityMixer2MainLoop: 
 
	ldmia	r0!, {r3-r4}			@ r3, r4 have 4 8bit vals 
	 
	conv8bitToQual2	r3, r5 
	conv8bitToQual2	r3, r6 
	conv8bitToQual2	r3, r7 
	conv8bitToQual2	r3, r8 
 
	conv8bitToQual2	r4, r9 
	conv8bitToQual2	r4, r10 
	conv8bitToQual2	r4, r11 
	conv8bitToQual2	r4, r12 
	 
	stmia	r1!, {r5-r12}			@ store samples 
 
	sub		r2, r2, #8 
	cmp		r2, #8 
	bgt		QualityMixer2MainLoop 
	 
QualityMixer2Remainder: 
	teq		r2, #0 
	beq		QualityMixerExit 
 
QualityMixer2RemainderLoop: 
 
	ldrb	r4, [r0] 
	add		r0, r0, #1 
	mov		r4, r4, LSL #8 
	orr		r4, r4, r4, LSL #16 
	str		r4, [r1] 
	add		r1, r1, #4	 
 
	subs	r2, r2, #1 
	bne		QualityMixer2RemainderLoop 
	b		QualityMixerExit 
 
 
/* *****************************************************************************
*/
 
QualityMixer4: 
	mov		r3, r2					@ r3 = size 
	and		r3, r3, #3				@ r3 = size % 4 (effectively get remainder when divided by 4) 
	 
	subs	r2, r2, r3				@ r2 = r2 - (r2 % 4) i.e. number of 4's 
	beq		QualityMixer4Remainder	@ only remainder case 
 
QualityMixer4MainLoop: 
 
	ldr		r4, [r0]				@ r4 has 4 8bit vals 
	add		r0, r0, #4				@ point to next 4 samples 
	 
	conv8bitToQual4 r4, r5, r6 
	conv8bitToQual4 r4, r7, r8 
	conv8bitToQual4 r4, r9, r10 
	conv8bitToQual4 r4, r11, r12 
	 
	stmia	r1!, {r5-r12}			@ store samples 
 
	subs	r2, r2, #4 
	bne		QualityMixer4MainLoop 
	 
QualityMixer4Remainder: 
	teq		r3, #0 
	beq		QualityMixerExit 
 
QualityMixer4RemainderLoop: 
 
	ldrb	r4, [r0] 
	add		r0, r0, #1 
	mov		r4, r4, LSL #8 
	orr		r4, r4, r4, LSL #16 
	mov		r5, r4 
	stmia	r1!, {r4-r5} 
	subs	r3, r3, #1 
	bne		QualityMixer4RemainderLoop 
	 
QualityMixerExit: 
 
	ldmia sp!, {r4-r12} 
	mov pc, lr 

